/* CWEIFFEL.DLL
 * ============
 *	Eiffel chromacoding, smartindenting, templates, etc 
 *  for Codewright, based on Premia's basic C. Tested with CW 3.0e/NT
 *	
 *	Adapted from Premia's bas.c by F. Arnaud (100041.375@compuserve.com)
 *
 *	version: 0.1 of 94-04-01
 *	todo: add/improve templates, check forgotten keywords for smart-indenting?
 */

/* bas.c copyright
 *
 *                  Copyright 1991,1992 Premia Corp.
 *                        All rights reserved
 *
 *    This source code is provided to the purchaser of the product
 *    for the purpose of allowing the purchaser to customize and/or
 *    extend the functionality of the product.  The source code   
 *    remains the property of Premia Corp. with all rights reserved.
 *    Any modifications to the source code are considered derivative
 *    works and all rights thereto are reserved to Premia Corp. except
 *    that the purchaser may use the modified product in the same
 *    permitted manner as the unmodified product.  It is expressly
 *    prohibited to distribute in any manner any part of the original
 *    source code, whether in source or object form, without the
 *    express written permission from Premia Corp.
 *
 *    Owner:
 *        Premia Corporation
 *        1075 NW Murray Blvd., Suite 268
 *        Portland OR 97229
 */
/* [premia has given its permission to distribute the derived work without
 * charge to codewright users]
 */

#include <windows.h>
#include <string.h>
#include <ctype.h>
#include <exports.h>
#include "cwstart.h"

#define MAX_BUF_LEN    100

static LPVOID _e_hash_names        = NIL;  /* pointer to the hash table containing */
                                           /* all 'Eiffel' keyword names. */

static LPVOID _e_hash_templates    = NIL;  /* pointer to the hash table containing */
                                           /* an abbreviated name and its template */

/* Define a structure to hold an abbreviation and the template to be
 * expanded when the abbreviation is entered into the buffer.
 * This structure will be the element in the hash table _e_hash_templates.
 */
typedef struct _LanguageTemplate {         
	LPSTR abbrev;
	LPSTR template;
} LanguageTemplate;


/* This string contains all keyword names which will be indented an extra level.
 */
LPSTR E_INDENT_STR ="\377from\377until\377loop" 
				"\377do\377when\377if\377then\377else\377elseif"
				"\377ensure\377rescue\377require\377local"
				"\377indexing\377feature\377creation\377inherit\377invariant"
				"\377rename\377export\377undefine\377redefine\377select"
				"\377";

/*
 * _e_templates[] contains the keyword abbreviation and the template for the
 * abbreviation.  To add a new name, insert it before the NIL entry or
 * call _e_assign_template().
 *
 * The special characters within the template are:
 *
 *     \n - new line
 *     &  - cursor position after template insertion
 *     @  - backspace
 *     \c - insert 'c' literally
 */

LanguageTemplate _e_templates[] = {
	{ "cla|clas|class", "class &\nend -- class" },
	{ "fea|feat|featu|featur", "feature" },
	{ "def|deff|deffe|deffer|deffere", "deffered" },
	{ "from", "from\n&\nuntil\n\nloop\n\nend" },
// more?
//	{ "ca|cas|case",                "case &:\n\tbreak;\n" },
//	{ "de|def|defa|defau|defaul|default",      
//                                    "default:\n\t&\nbreak;\n" },
//	{ "el|els|else",                "else\n{\n\t&\n@}" },
//	{ "ei",                         "else if (&)\n{\n}" },
//	{ "re|ret|retu|retur|return",   "return(&);\n" },
	{ NIL,       NIL },
 
};

/* Eiffel 3 keywords for chromacoding from ETL appendix G.2
 *	(standard types not included, will be colored as TYPES with the upper-case rule)
 */

LPSTR _e_names[] = {
"alias",	"all",		"and",		"as",		"check",	"class",
"creation",	"Current",	"debug",	"deferred",	"do",
"else",		"elseif",	"end",		"ensure",	"expanded",	"export",
"external",	"false",	"feature",	"from",		"frozen",	"if",
"implies",	"indexing",	"infix",	"inherit",	"inspect",	"invariant",
"is",		"like",		"local",	"loop",		"not",		"obsolete",
"old",		"once",		"or",		"prefix",	"redefine",	"rename",
"require",	"rescue",	"Result",	"retry",	"select",	"separate",
"strip",	"then",		"true",		"undefine",	"unique",	"until",
"variant",	"when",		"xor", 
NIL };

static long _e_in_comment( long offset );

void DLL _init( void );
BOOL DLL _e_expand( void );
void DLL _e_language_colors_update( long firstline, long lastline );
void DLL _e_language_colors( long numlines );
void DLL _e_assign_template( LPSTR abbrev, LPSTR template );
void DLL _e_add_keyword( LPSTR keyword );
BOOL DLL _e_indent( void );

////////////////////////////////////////////////////////////////////////////
// This section contains sample source code to be used with the Microsoft 'C'
// compiler.  No SDK is required unless a resource file is to be used.
// The functions in the SDK can be accessed by linking with a library
// created by running ilib.exe on the USER.EXE and GDI.EXE in your
// Window's System directory.

LIBMAIN    /* defines the LibMain() function for Borland C */


/* When this DLL is first loaded either by LibPreLoad() or LibAutoLoad(),
 * the function _init() is invoked.  Within this function, all DLL specific
 * initialization should be performed such as:
 *
 *         - Defining any module global keystrokes or keymaps to be used
 *         - Exporting any functions to be accessed through LibFunctionExec()
 *         - Autoloading any other DLLs.
 *
 */

void DLL 
_init()
{
	// Place all 'LibExport' calls here.
	// i.e.
	//   LibExport( "int myfunc LPSTR DWORD" );
	//
	static BOOL firstTime = TRUE;
	int    i;

	LibExport( "BOOL _e_expand");
	LibExport( "_e_assign_template LPSTR LPSTR");
	LibExport( "_e_add_keyword LPSTR");
	LibExport( "BOOL _e_indent");
	LibExport( "_e_language_colors_update long long");
	LibExport( "_e_language_colors long");

	/* Create a hash table to hold all of the 'Eiffel' keyword names.
	 */
	_e_hash_names = HashCreateTable( 41 /* a prime # */, sizeof( LPSTR ), 0 );
	i = 0;

	while ( _e_names[ i ] != NIL )
	{
		LPSTR XFAR *lpf = (LPSTR XFAR *)HashGetEntry( _e_hash_names, 
			_e_names[i], strlen( _e_names[i] ) );
		if (lpf)
		{
			*lpf = StrNew( _e_names[i] );
			i++;
		}
	}

	/* Create a hash table to hold all of the 'Eiffel' template names.
	 */
	_e_hash_templates = HashCreateTable( 57 /* a prime # */, sizeof( LPSTR ), 0 );
	i = 0;

	while ( _e_templates[ i ].abbrev != NIL )
	{
		_e_assign_template( _e_templates[i].abbrev, _e_templates[i].template );
		i++;
	}

	CompilerNewExt( ".e" );
}

 /*
 ** _e_expand()
 *
 *  DESCRIPTION:
 *    Attempts to perform 'Eiffel' language template expansion.  This function
 *    is normally bound to the <SPACE> key.  When invoked, it reads the 
 *    previous word.  If the word is in the list of defined templates,
 *    the template is extracted and inserted.  Some characters in the 
 *    templates have special meanings. They are:
 *
 *       \n - new line
 *       &  - cursor position after template insertion
 *       @  - backspace
 *       \  - take next character literally
 */
BOOL DLL
_e_expand( void )
{
	LPVOID htEntry;
	LPSTR  _e_string = NULL;
	LPSTR  str;

	MarkSavePos();

	/* Check to see if we are currently positioned inside a 'Eiffel' comment
	 */
	if (_e_in_comment( -1L ) != -1)
		goto _e_expand_incomplete;

	str = BufReadStr( 0x7FFFFFFF );     /* read entire line */

	_e_string = StrTrim( str, " \t" );
	StrFree( str );

	if (!_e_string)
		goto _e_expand_incomplete;

	/* Search for a keyword match in the hash table.
	 */
	htEntry = HashFindEntry( _e_hash_templates, _e_string, strlen( _e_string ) );
	StrFree( _e_string );

	if (htEntry)
	{
		/* A template exists for the current keyword, delete the keyword
		 * and insert the template.
		 */
		MarkDropPos( 1 );
		str = *(LPSTR*)htEntry;
		BufDelToEOL();

		_ext_expand_template( str );

	} 
	else
		goto _e_expand_incomplete;
	return TRUE;

_e_expand_incomplete:
	MarkRestorePos();

	return FALSE;
}

/*
 ** _e_language_colors_update()
 *
 *  This function does an intermediate update of the colors within
 *  a range of line numbers (firstline and lastline).  It must also
 *  determine if the first and last line specified are within a
 *  comment and act appropriately.
 */
void DLL
_e_language_colors_update( long firstline, long lastline )
{
	long startoff;

	if (firstline && lastline)
	{
		/* Begin coloring at the previous comment if
		 * currently located in the middle of one.
		 */
		_PosInit( 0 );
		if (firstline == 1)
			_PosPrevLine( 0L );
		else
			_PosNextLine( firstline - 1);
		startoff = _PosQOffset();

		/* Clear all attributes for the line range and
		 * the reset them within _e_language_colors.
		 */
		AttrSetColor( firstline, lastline, 1L, 0x7FFFFFFFL, 0 );
		_PosInit( startoff );
		_e_language_colors( lastline - firstline + 1 );
	}
}


/*
 ** _e_language_colors()
 *
 * If numlines == -1, start from the top of the buffer and encode all of them,
 * otherwise start at the current location of the number of lines specified.
 *
 * It is assumed that if numlines != -1, that _PosInit() has been called
 * and numlines is relative to the current position.
 */
void DLL
_e_language_colors( long numlines )
{
	long  startoff;
	long  startline;
	long  endoff;
	long  inc;
	WORD  ch;
	LPSTR buf;
	int   i = 0;
	int   comment_color = ColorComments( -1 );
	int   keyword_color = ColorKeywords( -1 );
	long  maxlines = ExtCommentSearchLimit( -1 );
	int   quote_color      = ColorAlternate1( -1 );
	int   number_color     = ColorAlternate2( -1 );
	int   type_color       = ColorAlternate3( -1 );
	int   preproc_color    = ColorAlternate4( -1 );


	buf = (LPSTR)MemAlloc( MAX_BUF_LEN );

	/* Start at the top of the buffer or current position 
	 */
	if (numlines == -1)
	{
		_PosInit( 0 );  
		numlines = 0x7FFFFFFF;
	}
	else
	{
		startoff  = _PosQOffset();
		startline = _PosQLine();
		endoff    = _e_in_comment( startoff );
		if (endoff == -1)
			_PosInit( startoff );
		else
		{
			_PosInit( endoff );
			numlines += (startline - _PosQLine());
		}
	}
	ch = _PosCurrentChar();
	startoff = _PosQOffset();

	maxlines = max( numlines, maxlines );

	while ((ch != EOF_CHAR) && (numlines > 0))
	{
		if (ch == '\n')
			numlines--;

		//Quoted
		if (ch == '"')
		{
			UINT quote = ch;

            startoff = _PosQOffset();

			do 
			{
				ch = _PosNextChar();
				if (ch == '\\')
				{
					ch = _PosNextChar();
					if ((ch == quote) || (ch == '\\'))
						ch = _PosNextChar();
				}
				if ((ch == EOF_CHAR) || (ch == '\n'))
					break;
			} while (ch != quote);

			if (ch == quote)
				ch = _PosNextChar();

			/* color the string */
			if (quote_color)
			{
				endoff = _PosQOffset();
	    		_PosInit( startoff );
				_PosSetColor( quote_color, endoff - startoff );
				_PosInit( endoff );
			}
			continue;
		}
		//Digit
		else if (number_color && isdigit( (BYTE)ch ))
		{
            int j = 0;
			UINT lastch = _PosPrevChar();

			if (lastch != EOF_CHAR)
				_PosNextChar();

			startoff = _PosQOffset();
			do
			{
                j++;
				ch = _PosNextChar();
			} while (isxdigit( (BYTE)ch ) || (ch == '.') || (ch == '_') || (  ((ch=='b') || (ch=='B')) && j == 1));

			if ((  lastch != ' ' 
				&& lastch != '\t' 
				&& lastch != '\r' 
				&& lastch != '\n' 
				&& lastch != EOF_CHAR
				&& !ispunct( (BYTE)lastch ))
				|| lastch == '_'
				|| (ch == '_'))
				continue;

            if (ch == 'l' || ch == 'L')
                ch = _PosNextChar();

			endoff = _PosQOffset();
    		_PosInit( startoff );
			_PosSetColor( number_color, endoff - startoff );
			_PosInit( endoff );
			continue;
		}
		//Alpha
		else if (isalpha( (BYTE)ch ) || (ch == '_'))
		{
			UINT lastch = _PosPrevChar();

			if (lastch != EOF_CHAR)
				_PosNextChar();

			i = 0;
			startoff = _PosQOffset();

			do
			{
				if (i < MAX_BUF_LEN)
					buf[i++] = (char)ch;
				ch = _PosNextChar();
			} while (isalpha( (BYTE)ch ) || isdigit( (BYTE) ch ) || (ch == '_'));
			/* identifiers may contain digits on !first position */

			if ((lastch != ' ' 
						&& lastch != '\t' 
						&& lastch != '\r' 
						&& lastch != '\n' 
						&& lastch != EOF_CHAR
						&& !ispunct( (BYTE)lastch ))
					|| (ch == '_'))
				i = 0;
			continue;
		}
		else if (i)
		{
			// keyword
			if (HashFindEntry( _e_hash_names, buf, i ))
			{
				endoff = _PosQOffset();
				_PosInit( startoff );
				_PosSetColor( keyword_color, endoff - startoff );
				_PosInit( endoff );
			}
			else
			// type
			{
				int ti;
				 
				for(ti=0;ti<i;ti++)
				{
					if(isalpha((BYTE)buf[ti]))
					{
						if(islower((BYTE)buf[ti]))
							break;
					}
				}
				if(ti==i) /* no alpha lowercase chars met => type name */
				{
					endoff = _PosQOffset();
					_PosInit( startoff );
					_PosSetColor( type_color, endoff - startoff );
					_PosInit( endoff );
				}
			}
			i = 0;
		} 

		// Comments '--'
		if (ch == '-')
		{
			UINT lastch = _PosPrevChar();

			if(lastch == '-')
			{
				startoff = _PosQOffset();

				inc = _PosNextLine( 1 );
				numlines--;
				endoff = _PosQOffset();
				_PosInit( startoff );
				_PosSetColor( comment_color, endoff - startoff - inc );
				_PosInit( endoff );
				ch = _PosCurrentChar();
				continue;
			}
			else if (lastch != EOF_CHAR)
				_PosNextChar();

		}

		ch = _PosNextChar();
	}
	MemFree( buf );
}


void DLL
_e_add_keyword( LPSTR keyword )
{
	if (keyword && *keyword)
	{
		LPSTR XFAR *lpf = (LPSTR XFAR *)HashGetEntry( _e_hash_names, 
				_fstrupr(keyword), strlen( keyword ) );
		if (lpf)
			*lpf = StrNew( keyword );
	}
}


/*
 ** _e_assign_template()
 *
 *  PARAMETERS:
 *    abbrev   - the keyword abreviation used to locate a template
 *    template - the encoded template string associated with 'abbrev'
 *
 *  DESCRIPTION:
 *    Provide a means of adding or modifying 'Eiffel' template strings
 *    without changing the code.
 */
void DLL
_e_assign_template( LPSTR abbrev, LPSTR template )
{
	_ext_assign_template( _e_hash_templates, abbrev, template );
}

/*
 ** _e_indent()
 *
 *  DESCRIPTION:
 *    Attempts to perform 'Eiffel' language smart indenting.  This function
 *    will behave differently based on the following conditions:
 *
 *    1. Within a comment - normal indentation
 *
 *    2. Not positioned at end of line - normal indentation
 *
 *    3. Current line prefixed by a name in E_INDENT_STR - perform smart
 *       indentation.  [removed: This may include skipping the next line if
 *       it consists only of an open brace.]
 */
BOOL DLL
_e_indent( void )
{
	long   curline;
	long   col;
	LPSTR  _e_string;
	LPSTR  str;
	long   matchpos;
//	WORD   endch;

	MarkSavePos();

	curline = BufQCurrentLine();

	/* Check to see if we are currently positioned inside a 'Eiffel' comment
	 */
	if (_e_in_comment( -1L ) != -1)
		goto _e_indent_incomplete;

	/* Check to see if there is a trailing '{' on the current line,
	 * and if so, indent on the next line.
	 */
//	do {
//		MovPrevChar( 1L );
//		endch = BufReadChar();
//	} while (endch == ' ' || endch == '\t');

	MarkRestorePos();

//	if (endch == '{')
//		goto _e_indent_next_line;

	MarkSavePos();

	/* Not in a comment so indent according to the current 'Eiffel' construct.
	 */
	if (!SrchFind( "^[ \t]*[a-zA-Z_]+\\c([ \t]|$)", SEARCH_REGEX , NIL )
			|| (BufQCurrentLine() != curline) )
		goto _e_indent_incomplete;

	/* Read the keyword at the beginning of the current line.
	 */
	col = BufQCurrentCol();
	SrchFind( "(^|[^a-zA-Z_])\\c[a-zA-Z_]", SEARCH_REGEX , NIL );
	str = BufReadStr( col - BufQCurrentCol() );
	_e_string = StrTrim( str, " \t" );
	StrFree( str );

	/* Set prefix and suffix to \377 to perform a quick lookup of the name
	 */
	str = StrNew( "\377" );
	_e_string = StrAppend( _e_string, str );
	str = StrAppend( str, _e_string );
	StrFree( _e_string );
	
	matchpos = StrMatch( str, E_INDENT_STR, SEARCH_FORWARD, NIL );
	StrFree( str );

	if (!matchpos)
		/* Not a 'Eiffel' keyword, do normal indentation. */
		goto _e_indent_incomplete;
		
	MarkRestorePos();
//	MarkSavePos();
//	/* Check to see if a '{' is by itself on the next line. If so, skip
//	 * it an indent on the line following it.
// 		 */
//	if (SrchFind( "^[ \t]*{\\c[ \t]*$", SEARCH_FORWARD | SEARCH_REGEX, NIL )
//			&& (BufQCurrentLine() == curline + 1) )
//		MarkDropPos( 1 );
//	else
//		MarkRestorePos();

//_e_indent_next_line:
	_ext_insert_indented_EOL();
	BufInsertChar( '\t' );
	return TRUE;

	_e_indent_incomplete:
		MarkRestorePos();

	return FALSE;
}



/*--  Private functions --*/


/*
 ** _e_in_comment()
 *
 *  Determines whether or not the current position in the buffer is
 *  located within a 'Eiffel' comment by search backward for a ' 
 *  comment declaration.
 *
 *  Returns -1 if not in a comment 
 *          >= 0 indicates the offset in the buffer where the comment begins
 */
static long
_e_in_comment( long offset )
{
	WORD  ch,lastch;

	/* Check to see if we are currently positioned inside a 'Eiffel' comment
	 */
	_PosInit( offset );
	ch = _PosCurrentChar();
	lastch = 0;

	do
	{
		if (ch == '\n')
			return -1;
		else if ((ch == '-') && (lastch == '-'))
		{
			return _PosQOffset();  /* currrent line is in a comment */
		}
		lastch=ch;
		ch = _PosPrevChar();
	} while (ch != EOF_CHAR);

	return -1L;
}

